///////////////////////////////////////////////////////////////////
// Debugging a pixel shader using "printf" written in HLSL.      //
// Written by Alexander Ehrath in 2006                           //
///////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////
//        debugging parameters (to be set up by programmer)          //
///////////////////////////////////////////////////////////////////

#define DebugStartU 0.2								// Specify the U coordinate of the location where you want to display the debug text
#define DebugStartV 0.5								// Specify the V coordinate of the location where you want to display the debug text

#define NumDigits 9									// Total number of digits to be displayed (including decimal point)
#define NumLines 3									// Total number of lines. Basically number of floats of interest.
#define DebugAlpha 0.8								// Opacity of Debug values over original texture (0.0-1.0)

#define FontScale 1.0								// Font size (1.0 = original size)
#define CharsPerLine 16								// Based on font texture having 16 characters per line
#define DecimalPoint 1								// Position of Decimal Point

///////////////////////////////////////////////////////////////////
//           definitions describing font texture used            //
///////////////////////////////////////////////////////////////////

#define InvFontScale (1.0 / FontScale)
#define CharSizeUV (1.0 / CharsPerLine)				// assuming characters are square in size (eg. 8x8 pixel characters)
#define InvCharSizeUV (CharsPerLine)				// assuming characters are square in size (eg. 8x8 pixel characters)
#define FirstDigitOffsetU CharSizeUV * 0			// U offset to character "0" in font set
#define FirstDigitOffsetV CharSizeUV * 1			// V offset to character "0" in font set
#define DebugAreaU CharSizeUV*NumDigits*FontScale	// Macro to compute total debugging area in U
#define DebugAreaV CharSizeUV*NumLines*FontScale	// Macro to compute total debugging area in V

///////////////////////////////////////////////////////////////////
//                debug font texture and sampler                 //
///////////////////////////////////////////////////////////////////

texture fontTexture;

sampler FontSampler = sampler_state
{
    Texture   = (fontTexture);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

///////////////////////////////////////////////////////////////////
//                          printf code                          //
///////////////////////////////////////////////////////////////////

float4 DebugValues(float2 texcoords, float4 color, float DebugArray[NumLines])
{
	float4 outcolor = color;
    if(texcoords.x < DebugAreaU+DebugStartU && texcoords.x > DebugStartU && texcoords.y < DebugAreaV+DebugStartV && texcoords.y > DebugStartV )
    {
        outcolor *= 1.0 - DebugAlpha;						// compute out color opacity
        
        float2 DebugUV = texcoords;							// Start with the textures incoming texture coordinates
        
        DebugUV.x -=  DebugStartU;							// subtract our debug region from it, to get into local debug UV space.
        DebugUV.y -=  DebugStartV;
        
		DebugUV *= InvFontScale;							// now scale these coordinates to define the size of our font
		
		int DigitSelect = DebugUV.x * InvCharSizeUV;		// figure out which digit the U coordinate falls in
		int LineSelect = DebugUV.y * InvCharSizeUV;			// figure out which line (i.e. which one of the debug values) the V coordinate falls in
		
		float debugVal = DebugArray[LineSelect];			// pull the value of interest out of the array of debug values
		
		DebugUV.x += FirstDigitOffsetU  - (CharSizeUV * DigitSelect);		
		DebugUV.y += FirstDigitOffsetV  - (CharSizeUV * LineSelect);
		bool negative = false;
		if(debugVal < 0)
		{
			negative = true;
			debugVal *= -1;
		}
		if(DigitSelect == 0)
		{
			if(negative)
			{
				DebugUV.x += CharSizeUV * 13;		// Negative sign is 13 characters to the right (relative to the character '0')
				DebugUV.y -= CharSizeUV * 1;		// Negative sign is 1 character up (relative to the character '0')
			}
			else
			{
				DebugUV.x += CharSizeUV * 11;		// Plus sign is 11 characters to the right (relative to the character '0')
				DebugUV.y -= CharSizeUV * 1;		// Plus sign is 1 character up (relative to the character '0')
			}
		}
		else if(DigitSelect == DecimalPoint)
		{
			DebugUV.x += CharSizeUV * 14;		// Decimal Point is 14 characters to the right (relative to the character '0')
			DebugUV.y -= CharSizeUV * 1;		// Decimal Point is 1 character up (relative to the character '0')
		}
		else 
		{
				if(DigitSelect > DecimalPoint)
				{
					debugVal = (int)(debugVal * pow(10,DigitSelect - DecimalPoint));
				}
				else
				{
					debugVal = (int)(debugVal *  pow(10,-DecimalPoint + DigitSelect + 1));
				}	
				
				if(DigitSelect != 0 || debugVal < 10.0)
				{
					float val = (int)(debugVal * 0.1);
					val *= 10;
					debugVal -= val;			
				}
				DebugUV.x += (CharSizeUV * debugVal);		
		}
        outcolor += tex2D( FontSampler, DebugUV ) * DebugAlpha;
    }
    return outcolor;
}

///////////////////////////////////////////////////////////////////
//                  regular shader definitions                   //
///////////////////////////////////////////////////////////////////

float4x4 worldViewProj : WorldViewProjection;
float4x4 worldView : WorldView;
float2 testCoordinate;
float3 lightDir;

texture decalTexture;
texture bumpTexture;

sampler ImageSampler = sampler_state
{
    Texture   = (decalTexture);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

sampler BumpSampler = sampler_state
{
    Texture   = (bumpTexture);
    MipFilter = LINEAR;
    MinFilter = LINEAR;
    MagFilter = LINEAR;
};

///////////////////////////////////////////////////////////////////
//                  vertex input definitions                     //
///////////////////////////////////////////////////////////////////

struct VS_INPUT
{
    float3 position    : POSITION;
	float3 normal      : NORMAL;
    float2 texture0	   : TEXCOORD0;
};

///////////////////////////////////////////////////////////////////
//                  vertex output definitions                    //
///////////////////////////////////////////////////////////////////

struct VS_OUTPUT
{
    float4 hposition : POSITION;
    float2 texture0  : TEXCOORD0;
    float2 texture1  : TEXCOORD1;
    float3 normal    : TEXCOORD2;
};

struct PS_OUTPUT
{
    float4 color : COLOR;
};

///////////////////////////////////////////////////////////////////
//                       vertex shader                           //
///////////////////////////////////////////////////////////////////

VS_OUTPUT vs( VS_INPUT IN )
{
    VS_OUTPUT OUT;

    OUT.hposition = mul( worldViewProj, float4(IN.position, 1) );

    OUT.texture0 = IN.texture0;
    OUT.texture1 = testCoordinate;
	OUT.normal = IN.normal;
    return OUT;
}


///////////////////////////////////////////////////////////////////
//              pixel shader to be debugged                      //
///////////////////////////////////////////////////////////////////

PS_OUTPUT ps( VS_OUTPUT IN )
{
    PS_OUTPUT OUT;
    
    // do normal bump mapping
	float4 viewNormal = mul(float4(IN.normal.xyz + tex2D( BumpSampler, IN.texture0 ),0.0),worldView);
    float shadecolor = dot(viewNormal.xyz, lightDir);
    OUT.color = tex2D( ImageSampler, IN.texture0 ) * shadecolor;		

    // set up a normal fetched from the bunp texture at a constant texture coordinate (passed in from C++ code)
	float4 debugnormal = normalize( mul(float4(IN.normal.xyz + tex2D( BumpSampler, IN.texture1 ),0.0),worldView));
	float DebugArray[NumLines];
    DebugArray[0] = debugnormal.x;
    DebugArray[1] = debugnormal.y;
    DebugArray[2] = debugnormal.z;
    				
	// mark the location of the debug pixel in pink
    float2 diff = IN.texture0 - IN.texture1;
    float len = sqrt(diff.x * diff.x + diff.y * diff.y);
    len = min (len*100.0, 1.0);
    OUT.color = lerp(float4(1.0f,0.0f,1.0f,1.0f), DebugValues(IN.texture0, OUT.color, DebugArray), len);
    return OUT;
}


technique DemoDebugTechnique
{
    pass Pass0
    {
        Lighting = FALSE;
        VertexShader = compile vs_3_0 vs();
        PixelShader  = compile ps_3_0 ps();
    }
}

